using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;

namespace Stopper {
	public class AdditionalControllers : MVRScript {
		// Prefabricated UI for joint controllers. Not sure how to get ahold of this
		// other than copying from somewhere
		GameObject controllerUIPrefab;
		
		// A sample controller for retrieving controller meshes. Again, not sure
		// how to find this other than copying from elsewhere
		FreeControllerV3 sampleController;
		
		Atom mainAtom;
		
		GameObject jointControl;
		GameObject ui;
		
		GameObject uiPlaceholder;
		
		FreeControllerV3 controller;
		
		// Rigidbodies to add controllers for. 
		List<string> rigidBodyControllerList = new List<string> { 
			"tongueTip",
			"rIndex3", "lIndex3",
			"rMid3", "lMid3",
			"rRing3", "lRing3",
			"rPinky3", "lPinky3",
			"rThumb3", "lThumb3",
		};  
		
		public override void Init() {
		}
		
		private GameObject FindGameObject(GameObject parent, string name){
			Transform childTransform = parent.transform.Find(name);
			
			if(childTransform == null){
				return null;
			} else {
				return childTransform.gameObject;
			}
		}
		
		private GameObject GetOrCreateChildObject(GameObject parent, string name) {
			GameObject gameObject = FindGameObject(parent, name);
			if(gameObject == null) {
				gameObject = new GameObject(name);
				gameObject.transform.parent = parent.transform;
			} 
			
			return gameObject;
		}
		
		private Component GetOrCreateComponent(GameObject parent, System.Type type) {
			Component component = parent.GetComponent(type) as Component;
			if(component == null){
				component = parent.AddComponent(type) as Component;
			}
			return component;
		}
		
		List<GameObject> GetAllChildren(GameObject parent, List<GameObject> objectList = null)
		{
			if (objectList == null) objectList = new List<GameObject>();
		  
			foreach (Transform child in parent.transform) {
				objectList.Add(child.gameObject);
				GetAllChildren(child.gameObject, objectList);
			}
			return objectList;
		}
		
		//-------Start------
		private void Start()
		{
			StartCoroutine("AddControllers");
		}
		
		private IEnumerator AddControllers()
		{
			/* 
				This script works by modifying the Person prefab - the template from which other People are created.
				
				Why? Because Atoms have an immutable freeControllers array that's used for various things like 'select
				collider from scene'. 
				
				If that doesn't matter to you, set mainAtom = containingAtom and delete all the prefab code.			
			*/
			
			// Check to see if the prefab already exists (there's a person in the scene)
			GameObject personPrefab = SuperController.singleton.GetCachedPrefab("a_per", "Person");

			
			if(personPrefab == null){
				// No people in the scene yet. Load one (without adding to scene) and register it as the prefab. 
				// This might have unintended consequences. 
				SuperController.LogMessage("Person prefab not cached. Loading from asset bundle.");
				AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync("a_per", "Person", typeof(GameObject));
				if (request == null)
				{
					SuperController.LogError("Unable to load Person prefab");
					yield break;
				}
				yield return base.StartCoroutine(request);

				personPrefab = request.GetAsset<GameObject>();
				if (personPrefab != null)
				{
					SuperController.singleton.RegisterPrefab("a_per", "Person", personPrefab);
					mainAtom = personPrefab.GetComponent<Atom>();
				} else {
					SuperController.LogError("Unable to load Person prefab");
					yield break;
				}
			} else {
				mainAtom = personPrefab.GetComponent<Atom>();
			}
			
			GameObject sampleHipControl = null;
			
			// Load an existing controller joint ui to use as a template. Not sure how else to do it.
			foreach (GameObject go in Resources.FindObjectsOfTypeAll(typeof(GameObject)) as GameObject[]){
				if(go.name =="UIFreeControllerJoint") {
					controllerUIPrefab = go;
					continue;
				}
			}
			
			foreach (GameObject go in GetAllChildren(mainAtom.gameObject)){
				if(go.name == "hipControl") {
					sampleHipControl = go;
					continue;
				}
				
				if(go.name == "JointControl") {
					jointControl = go;
					continue;
				}
				
				if(go.name == "UI" && go.transform.parent.gameObject == mainAtom.gameObject) {
					ui = go;
					continue;
				}
			}
			sampleController = sampleHipControl.GetComponent<FreeControllerV3>();
			
			foreach(Rigidbody rb in mainAtom.gameObject.GetComponentsInChildren<Rigidbody>()){
				//if(rb.name.Contains("tongue")){
				//rb.solverVelocityIterations = 10;
				//}
				
				if(!rigidBodyControllerList.Contains(rb.name)){
					continue;
				}
		
				
				rb.gameObject.SetActive(false);
				
				// ForceReceiver
				ForceReceiver forceReceiver = GetOrCreateComponent(rb.gameObject, typeof(ForceReceiver)) as ForceReceiver;
				forceReceiver.containingAtom = mainAtom;
				
				// Parent object for controls
				//  Needs a Rigidbody, SphereCollider, FreeControllerV3, and a joint tying it to the target Rigidbody
				GameObject control = GetOrCreateChildObject(jointControl, rb.name+"Control");
				control.SetActive(false);
				control.layer = 8;

				Rigidbody controlRB = GetOrCreateComponent(control, typeof(Rigidbody)) as Rigidbody;
				controlRB.isKinematic = false;
				controlRB.useGravity = false;
				controlRB.detectCollisions = true;
				
				ConfigurableJoint controlJoint = rb.gameObject.AddComponent(typeof(ConfigurableJoint)) as ConfigurableJoint;
				controlJoint.connectedBody = controlRB;
				controlJoint.autoConfigureConnectedAnchor = false;
				
				FreeControllerV3 controller = GetOrCreateComponent(control, typeof(FreeControllerV3)) as FreeControllerV3;
				controller.containingAtom = mainAtom;
				controller.followWhenOffRB = rb;
				controller.useGravityOnRBWhenOff = false;
				controller.meshScale = 0.3f;
				controller.enableSelectRoot = true;
				controller.enabled = true;
				controller.hidden = false;

				controller.material = sampleController.material;
				if(controller.onPositionMesh == null){
					controller.onPositionMesh = sampleController.onPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.onPositionMesh, controller.transform);
					controller.offPositionMesh = sampleController.offPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.offPositionMesh, controller.transform);
					controller.followingPositionMesh = sampleController.followingPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.followingPositionMesh, controller.transform);
					controller.holdPositionMesh = sampleController.holdPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.holdPositionMesh, controller.transform);
					controller.lockPositionMesh = sampleController.lockPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.lockPositionMesh, controller.transform);
					controller.onRotationMesh = sampleController.onRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.onRotationMesh, controller.transform);
					controller.offRotationMesh = sampleController.offRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.offRotationMesh, controller.transform);
					controller.followingRotationMesh = sampleController.followingRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.followingRotationMesh, controller.transform);
					controller.holdRotationMesh = sampleController.holdRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.holdRotationMesh, controller.transform);
					controller.lockRotationMesh = sampleController.lockRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.lockRotationMesh, controller.transform);
					controller.lookAtRotationMesh = sampleController.lookAtRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.lookAtRotationMesh, controller.transform);
					controller.moveModeOverlayMesh = sampleController.moveModeOverlayMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.moveModeOverlayMesh, controller.transform);
					controller.rotateModeOverlayMesh = sampleController.rotateModeOverlayMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.rotateModeOverlayMesh, controller.transform);
					controller.deselectedMesh = sampleController.deselectedMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.deselectedMesh, controller.transform);
				}
				
				SphereCollider collider = GetOrCreateComponent(control, typeof(SphereCollider)) as SphereCollider;
				collider.enabled = true;
				collider.radius = 0.1f;
				
				// Controller Link. Not sure what this is used for.
				GameObject controlLink = GetOrCreateChildObject(jointControl, rb.name+"Link");
				controlLink.layer = 8;
				
				SphereCollider linkCollider = GetOrCreateComponent(controlLink, typeof(SphereCollider)) as SphereCollider;
				linkCollider.enabled = true;
				linkCollider.radius = 0.1f;
				
				Rigidbody linkRB = GetOrCreateComponent(controlLink, typeof(Rigidbody)) as Rigidbody;
				linkRB.isKinematic = true;
				linkRB.useGravity = false;
				linkRB.detectCollisions = false;
				
				FreeControllerV3Link freeControllerLink = GetOrCreateComponent(controlLink, typeof(FreeControllerV3Link)) as FreeControllerV3Link;
				freeControllerLink.linkedController = controller;
				
				// UI				
				uiPlaceholder = GetOrCreateChildObject(ui, "UIPlaceholder"+rb.name );
				uiPlaceholder.SetActive(false);
				
				MoveAndRotateAsHUDAnchor moveAsHud = GetOrCreateComponent(uiPlaceholder, typeof(MoveAndRotateAsHUDAnchor)) as MoveAndRotateAsHUDAnchor;
				moveAsHud.scaleAsAnchor = true;
				moveAsHud.anchorNum = HUDAnchor.AnchorNum.Two;
									
				UIConnectorMaster connectorMaster = GetOrCreateComponent(uiPlaceholder, typeof(UIConnectorMaster)) as UIConnectorMaster;
				connectorMaster.UIPrefab = controllerUIPrefab.transform;
						
				UIConnector connector = GetOrCreateComponent(uiPlaceholder, typeof(UIConnector)) as UIConnector;
				connector.receiver = controller;
				connector.receiverTransform = control.transform;
				
				controller.UITransforms = new Transform[]{uiPlaceholder.transform};
				controller.UITransformsPlayMode = new Transform[]{};
				
				control.SetActive(true);
				uiPlaceholder.SetActive(true);
				rb.gameObject.SetActive(true);
				
				controller.currentPositionState = FreeControllerV3.PositionState.Off;
				controller.currentRotationState = FreeControllerV3.RotationState.Off;	
				controller.guihidden = true;
			}
		}
		
		private void OnDisable()
		{
			// Clean up maybe? Seems like a hassle.
		}
	}
}